from threading import _RLock, _allocate_lock

from dank_mids._logging import getLogger


logger = getLogger("dank_mids.AlertingRLock")
acquire_lock = _RLock.acquire


class AlertingRLock(_RLock):  # type: ignore [misc]
    __slots__ = "_block", "_owner", "_count", "_name"

    def __init__(self, name: str) -> None:
        """
        Initializes the reentrant lock with a given name.

        The name can be used for debugging and logging purposes to track lock usage.

        Examples:
            >>> lock = AlertingRLock(name="example")
            >>> lock._name
            'example'

        See Also:
            :class:`UIDGenerator`
        """
        self._block = _allocate_lock()
        self._owner = None
        self._count = 0
        self._name = name

    def acquire(self, blocking: bool = True, timeout: float = -1.0) -> bool:  # type: ignore [override]
        """
        Attempts to acquire the lock with an initial fixed timeout.

        This method first attempts to acquire the lock in a non-blocking manner with a timeout
        of 5 seconds. If this initial attempt is unsuccessful, a warning message is logged and
        a secondary acquisition attempt is made using the provided `blocking` and `timeout`
        parameters. Note that the return value of the secondary attempt is not captured or returned;
        this method always returns the result of the initial lock acquisition attempt.

        Parameters:
            blocking: Whether to block during the secondary acquisition attempt.
            timeout: The maximum time to wait during the secondary acquisition attempt.

        Returns:
            bool: The result from the initial lock acquisition attempt.

        Examples:
            >>> lock = AlertingRLock("example")
            >>> # Attempt to acquire the lock. If the initial 5-second attempt fails,
            >>> # a warning is logged and a secondary attempt is made.
            >>> result = lock.acquire(blocking=True, timeout=10)
            >>> isinstance(result, bool)
            True

        See Also:
            :meth:`_RLock.acquire`
        """
        acquired = acquire_lock(self, blocking=False, timeout=5)
        if not acquired:
            # NOTE: I havent seen this log in a very long time, maybe we should remove the lock soon
            logger.warning("wtf?! %s with name %s is locked!", self, self._name)
            acquire_lock(self, blocking=blocking, timeout=timeout)
        return acquired


__all__ = ["AlertingRLock"]



from contextlib import suppress

from web3.eth import AsyncEth

from dank_mids.helpers import setup_dank_w3_from_sync
from dank_mids.brownie_patch.types import (
    DankContractCall,
    DankContractMethod,
    DankContractTx,
    DankOverloadedMethod,
)

__all__ = ["DankContractCall", "DankContractMethod", "DankContractTx", "DankOverloadedMethod"]

from dank_mids.eth import DankEth
from dank_mids.helpers._helpers import DankWeb3

dank_web3: DankWeb3
"""
An instance of :py:class:`~DankWeb3`, which is a wrapper around brownie’s standard Web3 object with Dank Middleware injected.
This is initialized if Brownie is installed and connected when this module is loaded.
If Brownie is not installed or not connected to an RPC, this instance will not be available.
"""

dank_eth: DankEth
"""
An instance of :py:class:`~DankEth`, providing access to Ethereum blockchain methods via :py:const:`dank_web3`.
This is initialized if Brownie is installed and connected when this module is loaded.
If Brownie is not installed or not connected to an RPC, this instance will not be available.
"""

# If using dank_mids with brownie, and brownie is connected when this file executes, you will get a 'dank_w3' async web3 instance with Dank Middleware here.
with suppress(ImportError):
    from brownie import network, web3

    if network.is_connected():
        from dank_mids.brownie_patch.contract import Contract, patch_contract

        dank_web3 = setup_dank_w3_from_sync(web3)
        dank_eth = dank_web3.eth
        __all__ += ["Contract", "patch_contract", "dank_web3", "dank_eth"]



from typing import Any, Final, final

from dank_mids import ENVIRONMENT_VARIABLES
from dank_mids._logging import getLogger


@final
class DummyLogger:
    """
    A dummy logger class that provides a no-op implementation of the info
    method of logging.Logger.

    This class is used to replace the actual logger when demo mode is not
    enabled, saving processing time by avoiding unnecessary logging operations.
    """

    @staticmethod
    def info(*args: Any, **kwargs: Any) -> None:
        """
        Does nothing. Mimics the API of logging.Logger.info by accepting
        any arguments.

        Args:
            *args: Positional arguments (ignored).
            **kwargs: Keyword arguments (ignored).
        """


# Choose between a real logger and a dummy logger based on the demo mode setting
if ENVIRONMENT_VARIABLES.DEMO_MODE:  # type: ignore [has-type]
    _logger = getLogger("dank_mids.demo")
else:
    _logger = DummyLogger()  # type: ignore [assignment]

demo_logger: Final = _logger



from asyncio import as_completed
from functools import wraps
from importlib.metadata import version
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Iterable,
    List,
    Literal,
    Optional,
    TypeVar,
    Union,
)

from eth_utils.curried import apply_formatter_if, apply_formatters_to_dict, apply_key_map, is_null
from eth_utils.toolz import assoc, complement, compose, merge
from hexbytes import HexBytes
from multicall.utils import get_async_w3
from typing_extensions import Concatenate, ParamSpec
from web3 import Web3
from web3._utils.rpc_abi import RPC
from web3.datastructures import AttributeDict
from web3.providers import HTTPProvider
from web3.providers.async_base import AsyncBaseProvider
from web3.providers.base import BaseProvider
from web3.types import Formatters, FormattersDict, RPCEndpoint, RPCResponse

from dank_mids.eth import DankEth
from dank_mids.types import AsyncMiddleware

if TYPE_CHECKING:
    from dank_mids._requests import JSONRPCBatch, Multicall

    Batch = TypeVar("Batch", "Multicall", "JSONRPCBatch")
else:
    Batch = Union["Multicall", "JSONRPCBatch"]

dank_w3s: List[Web3] = []
"""
A list that stores instances of :class:`Web3` objects that have been set up with Dank Middleware.
This list is used to keep track of all Dank-enabled Web3 instances created in the application.
"""

sync_w3s: List[Web3] = []
"""
A list that stores instances of synchronous Web3 objects.
This list is used to keep track of all synchronous Web3 instances that have been created or modified.
"""

T = TypeVar("T")
"""Generic type variable for use in type hints."""

P = ParamSpec("P")
"""Parameter specification for use in type hints."""

w3_version_major = int(version("web3").split(".")[0])
"""
The major version number of the currently installed web3.py package.
This is extracted from the full version string and stored as an integer for easy comparison.

Example: for `web3==6.0.1`, the major version is 6.
"""


class DankWeb3:
    """This is just a helper for type checkers. Your object will just be a modified :class:`~web3.Web3` object."""

    eth: DankEth


skip_poa_middleware = {
    1,  # eth mainnet
}
"""
By default, we include the poa middleware on all chains to prevent issues. 
But when we don't need it, we can remove it by adding the chainid to this set.
"""


def setup_dank_w3(async_w3: Web3) -> DankWeb3:
    """
    Sets up a DankWeb3 instance from a given Web3 instance.

    Args:
        async_w3: The Web3 instance to be wrapped.

    Returns:
        A new DankWeb3 instance with Dank Middleware injected.
    """
    assert async_w3.eth.is_async and isinstance(async_w3.provider, AsyncBaseProvider)
    # NOTE: We use this lookup to prevent errs where 2 project dependencies both depend on dank_mids and eth-brownie.
    if async_w3 not in dank_w3s:
        # NOTE: We import here to prevent a circular import
        from dank_mids.middleware import dank_middleware

        async_w3.middleware_onion.inject(dank_middleware, layer=0)
        if _sync_w3_from_async(async_w3).eth.chain_id not in skip_poa_middleware:
            async_w3.middleware_onion.add(geth_poa_middleware)
        dank_w3s.append(async_w3)
    async_w3.eth = DankEth(async_w3)
    return async_w3


def setup_dank_w3_from_sync(sync_w3: Web3) -> DankWeb3:
    """
    Sets up a DankWeb3 instance from a given synchronous Web3 instance.

    Args:
        sync_w3: The synchronous Web3 instance to be wrapped.

    Returns:
        A new DankWeb3 instance with Dank Middleware injected, supporting batched asynchronous operations.
    """
    assert not sync_w3.eth.is_async and isinstance(sync_w3.provider, BaseProvider)
    if sync_w3 not in sync_w3s:
        if w3_version_major >= 6:
            # this was default in prior versions but now optional. We want it.
            __add_sync_attrdict_middleware(sync_w3)
        if sync_w3.eth.chain_id not in skip_poa_middleware:
            __add_sync_poa_middleware(sync_w3)
        sync_w3s.append(sync_w3)
    return setup_dank_w3(get_async_w3(sync_w3))


def set_done(
    fn: Callable[Concatenate[Batch, P], Awaitable[T]],
) -> Callable[Concatenate[Batch, P], Awaitable[T]]:
    """
    A decorator that sets the '_done' flag of a _Request object after the decorated function completes.

    Args:
        fn: The function to be decorated.

    Returns:
        A wrapped version of the input function that sets the '_done' flag.
    """

    @wraps(fn)
    async def set_done_wrap(self: Batch, *args: P.args, **kwargs: P.kwargs) -> T:
        retval = await fn(self, *args, **kwargs)
        self._done.set()
        return retval

    return set_done_wrap


# Everything below is in web3.py now, but dank_mids currently needs a version that predates them.

is_not_null = complement(is_null)
"""
A function that returns True if the input is not null, False otherwise.
This is the complement of the `is_null` function.
"""

FORMATTER_DEFAULTS: FormattersDict = {
    "request_formatters": {},
    "result_formatters": {},
    "error_formatters": {},
}
"""
Default formatters dictionary used in middleware configuration.
It provides empty dictionaries for request, result, and error formatters.
"""

remap_geth_poa_fields = apply_key_map(
    {
        "extraData": "proofOfAuthorityData",
    }
)
"""
A function that remaps Geth Proof of Authority (PoA) fields.
It changes the key 'extraData' to 'proofOfAuthorityData'.
"""

pythonic_geth_poa = apply_formatters_to_dict(
    {
        "proofOfAuthorityData": HexBytes,
    }
)
"""
A function that applies formatters to Geth Proof of Authority (PoA) fields.
It converts the 'proofOfAuthorityData' field to HexBytes.
"""

geth_poa_cleanup = compose(pythonic_geth_poa, remap_geth_poa_fields)
"""
A composed function that applies both remapping and formatting to Geth PoA fields.
It first remaps the fields and then applies the pythonic formatters.
"""


async def geth_poa_middleware(
    make_request: Callable[[RPCEndpoint, Any], Any], w3: Web3
) -> AsyncMiddleware:
    middleware = await async_construct_formatting_middleware(
        result_formatters={
            RPC.eth_getBlockByHash: apply_formatter_if(is_not_null, geth_poa_cleanup),
            RPC.eth_getBlockByNumber: apply_formatter_if(is_not_null, geth_poa_cleanup),
        },
    )
    return await middleware(make_request, w3)  # type: ignore [arg-type, return-value]


async def async_construct_formatting_middleware(
    request_formatters: Optional[Formatters] = None,
    result_formatters: Optional[Formatters] = None,
    error_formatters: Optional[Formatters] = None,
) -> AsyncMiddleware:
    async def ignore_web3_in_standard_formatters(
        _w3: "Web3",
        _method: RPCEndpoint,
    ) -> FormattersDict:
        return dict(
            request_formatters=request_formatters or {},
            result_formatters=result_formatters or {},
            error_formatters=error_formatters or {},
        )

    return await async_construct_web3_formatting_middleware(  # type: ignore [return-value]
        ignore_web3_in_standard_formatters
    )


async def async_construct_web3_formatting_middleware(
    async_web3_formatters_builder: Callable[
        ["Web3", RPCEndpoint], Coroutine[Any, Any, FormattersDict]
    ],
) -> Callable[[Callable[[RPCEndpoint, Any], Any], "Web3"], Coroutine[Any, Any, AsyncMiddleware]]:
    async def formatter_middleware(
        make_request: Callable[[RPCEndpoint, Any], Any],
        async_w3: "Web3",
    ) -> AsyncMiddleware:
        async def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
            formatters = merge(
                FORMATTER_DEFAULTS,
                await async_web3_formatters_builder(async_w3, method),
            )
            request_formatters = formatters.pop("request_formatters")

            if method in request_formatters:
                formatter = request_formatters[method]
                params = formatter(params)
            response = await make_request(method, params)

            return _apply_response_formatters(method=method, response=response, **formatters)

        return middleware

    return formatter_middleware


def _apply_response_formatters(
    method: RPCEndpoint,
    response: RPCResponse,
    result_formatters: Formatters,
    error_formatters: Formatters,
) -> RPCResponse:
    def _format_response(
        response_type: Literal["result", "error"],
        method_response_formatter: Callable[..., Any],
    ) -> RPCResponse:
        appropriate_response = response[response_type]
        return assoc(response, response_type, method_response_formatter(appropriate_response))

    if "result" in response and method in result_formatters:
        return _format_response("result", result_formatters[method])
    elif "error" in response and method in error_formatters:
        return _format_response("error", error_formatters[method])
    else:
        return response


def _sync_w3_from_async(w3: Web3) -> Web3:
    """
    Creates a synchronous Web3 instance from an asynchronous one.

    This function is used internally to create a sync Web3 instance
    for operations that require synchronous execution.

    Args:
        w3: An asynchronous Web3 instance.

    Returns:
        A synchronous Web3 instance with the same endpoint URI.

    Raises:
        ValueError: If the input Web3 instance is not asynchronous.
    """
    if not w3.eth.is_async or not isinstance(w3.provider, AsyncBaseProvider):
        raise ValueError("Dank Middleware can only be applied to an asynchronous Web3 instance.")
    sync_provider = HTTPProvider(w3.provider.endpoint_uri)
    sync_w3: Web3 = Web3(provider=sync_provider)
    # We can't pickle middlewares to send to process executor.
    # The call has already passed thru all middlewares on the user's Web3 instance.
    sync_w3.middleware_onion.clear()
    sync_w3.provider.middlewares = ()
    return sync_w3


def _make_hashable(obj: Any) -> Any:
    """
    Converts an object into a hashable type if possible.

    This function is used internally to ensure that objects can be used as
    dictionary keys or in sets.

    Args:
        obj: The object to make hashable.
    """
    if isinstance(obj, (list, tuple)):
        return tuple(map(_make_hashable, obj))
    elif isinstance(obj, dict):
        return AttributeDict(zip(obj.keys(), map(_make_hashable, obj.values())))  # type: ignore [arg-type]
    return obj


def __add_sync_attrdict_middleware(sync_w3: Web3) -> None:
    """
    Add the attrdict middleware to a synchronous Web3 instance.

    Args:
        sync_w3: The synchronous Web3 instance to add the middleware to.

    Raises:
        ValueError: If the provided Web3 instance is asynchronous.
    """
    # we do the import here because it is only present in web3 >= 6
    from web3.middleware import attrdict_middleware

    if sync_w3.eth.is_async:
        raise ValueError("You must pass in a sync Web3 instance.")
    try:
        sync_w3.middleware_onion.add(attrdict_middleware)
    except ValueError as e:
        if str(e) != "You can't add the same un-named instance twice":
            raise
        # The middleware is already added


def __add_sync_poa_middleware(sync_w3: Web3) -> None:
    """
    Add the Proof of Authority (PoA) middleware to a synchronous Web3 instance.

    This ensures compatibility with PoA chains.

    Args:
        sync_w3: The synchronous Web3 instance to add the middleware to.
    """
    if w3_version_major >= 7:
        from web3.middleware.proof_of_authority import ExtraDataToPOAMiddleware

        sync_w3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)
    elif w3_version_major >= 6:
        from web3.middleware import geth_poa_middleware

        sync_w3.middleware_onion.inject(geth_poa_middleware, layer=0)



import re
from functools import cached_property
from logging import getLogger
from time import time
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Coroutine,
    Dict,
    Final,
    List,
    Literal,
    NewType,
    Optional,
    Set,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
    Union,
    overload,
)

import evmspec
from dictstruct import DictStruct
from eth_typing import ChecksumAddress
from evmspec.data import Address, BlockNumber, ChainId, Wei, uint, _decode_hook
from evmspec.structs.block import BaseBlock, Block, MinedBlock, ShanghaiCapellaBlock
from evmspec.structs.log import Log
from hexbytes import HexBytes
from msgspec import UNSET, Raw, ValidationError, field
from msgspec.json import Decoder
from msgspec.json import decode as json_decode
from web3.datastructures import AttributeDict
from web3.types import RPCEndpoint, RPCError, RPCResponse

from dank_mids import constants, stats
from dank_mids._exceptions import (
    BadResponse,
    ChainstackRateLimitError,
    ExceedsMaxBatchSize,
    ExecutionReverted,
    OutOfGas,
    PayloadTooLarge,
    QuiknodeRateLimitError,
)
from dank_mids.helpers._codec import RawResponse, decode_string, encode
from dank_mids.helpers.lru_cache import lru_cache_lite_nonull

if TYPE_CHECKING:
    from dank_mids._requests import Multicall


T = TypeVar("T")


BlockId = NewType("BlockId", str)
"""A type representing the identifier for a specific block in the blockchain."""

BatchId = Union[int, str]
"""A type representing the identifier for a batch of operations, which can be either an integer or a string."""

Multicalls = Dict[BlockId, "Multicall"]
"""
A dictionary mapping BlockId to Multicall objects.

See Also:
    :class:`dank_mids._requests.Multicall`: The Multicall class used in this mapping.
"""

eth_callParams = TypedDict("eth_callParams", {"to": ChecksumAddress, "data": str})
"""
A typed dictionary representing the parameters for an eth_call.

See Also:
    :meth:`web3.eth.Eth.call`: Web3's method to perform a call without creating a transaction.
"""

OverrideParams = TypedDict("OverrideParams", {"code": str})
"""A typed dictionary representing override parameters."""

JsonrpcParams = List[Union[eth_callParams, BlockId, OverrideParams]]
"""A list of parameters for JSON-RPC calls, which should be eth_callParams, BlockId, and OverrideParams, in that order."""


# This type alias was introduced in web3 v5.28.0 but we like loose deps here so we recreate instead of import.
AsyncMiddleware = Callable[[RPCEndpoint, Any], Coroutine[Any, Any, RPCResponse]]
"""A type alias for asynchronous middleware functions."""

_list_of_stuff = List[Union[str, None, dict, list]]
"""A type alias for a list that can contain strings, None, dictionaries, or lists."""

_dict_of_stuff = Dict[str, Union[str, None, _list_of_stuff, Dict[str, Optional[Any]]]]
"""A type alias for a dictionary with string keys and values that can be strings, None, :obj:`_list_of_stuff`, or dictionaries with optional values."""

_nested_dict_of_stuff = Dict[str, Union[str, None, _list_of_stuff, _dict_of_stuff]]
"""A type alias for a nested dictionary structure."""


decode_nested_dict: Final = Decoder(type=_nested_dict_of_stuff).decode


class PartialRequest(DictStruct, frozen=True, omit_defaults=True, repr_omit_defaults=True):  # type: ignore [call-arg]
    """
    Represents a partial JSON-RPC request.

    While technially part of a request, we can successfully make requests to many nodes without including the `jsonrpc` field.

    This class leaves off the `jsonrpc` field reduce encoding burden and web traffic.

    This works with many but not all nodes.
    """

    method: str
    """The RPC method to be called."""

    id: Union[str, int]
    """The request identifier."""

    params: Optional[list] = None
    """The parameters for the RPC method call."""

    @property
    def data(self) -> bytes:
        return encode(self)


class Request(PartialRequest):
    """
    Represents a complete JSON-RPC request.

    Inherits from PartialRequest and adds the JSON-RPC version.
    """

    jsonrpc: Literal["2.0"] = "2.0"
    """The JSON-RPC version, always set to "2.0"."""

    @property
    def data(self) -> bytes:
        # we have to do some hacky stuff because omit_defaults kwarg on PartialRequest
        # is preventing jsonrpc field from being included in the encoded bytes
        return encode(self)[:-1] + b',"jsonrpc":"2.0"}'


class RPCErrorWithContext(RPCError):
    dank_mids_added_context: PartialRequest


_getattribute = object.__getattribute__


class ChainstackRateLimitContext(DictStruct, frozen=True):  # type: ignore [call-arg]
    """Chainstack doesn't use status code 429 for rate limiting, they attach one of these objects to a 200 response."""

    logger: ClassVar = getLogger("dank_mids.chainstack")
    _try_again_in: str = field(name="try_again_in")

    @property
    def try_again_in(self) -> float:
        """
        Calculates the time to wait before retrying the request.

        Returns:
            The number of seconds to wait before retrying.

        Raises:
            NotImplementedError: If the time string is in an unrecognized format.
        """
        error: Error = self.response.error
        decimal_string = error.data.try_again_in
        if "ms" in decimal_string:
            ms = float(decimal_string[:-2])
            self.logger.warning("rate limited by chainstack, retrying in %sms", ms)
            return ms / 1000
        elif "µs" in decimal_string:
            µs = float(decimal_string[:-2])
            self.logger.warning("rate limited by chainstack, retrying in %sµs", µs)
            return µs / 1000000
        raise NotImplementedError(f"must define a handler for decimal_string {decimal_string}")


class Error(DictStruct, frozen=True, omit_defaults=True, repr_omit_defaults=True):  # type: ignore [call-arg]
    """
    Represents an error in a JSON-RPC response.
    """

    code: int
    """The error code."""

    message: str
    """The error message."""

    # evm spec
    data: Optional[Union[str, ChainstackRateLimitContext]] = UNSET  # type: ignore [assignment]
    """
    Additional error data, if any.
    
    EVM specs say it should be a string, but some providers will return a dictionary here with even more context.
    """

    @overload
    def to_dict(self, *, context: None = None) -> RPCError: ...
    @overload
    def to_dict(self, *, context: PartialRequest) -> RPCErrorWithContext: ...
    def to_dict(self, *, context: Optional[PartialRequest] = None) -> RPCError:
        data = _getattribute(self, "data")
        if context is None:
            if data is UNSET:
                return RPCError(code=self.code, message=self.message)
            else:
                return RPCError(code=self.code, message=self.message, data=data)
        elif data is UNSET:
            return RPCErrorWithContext(
                code=self.code, message=self.message, dank_mids_added_context=context
            )
        else:
            return RPCErrorWithContext(
                code=self.code, message=self.message, data=data, dank_mids_added_context=context
            )


# some devving tools that will go away eventually
_dict_responses: Set[str] = set()
_str_responses: Set[str] = set()


_RETURN_TYPES = {
    "eth_call": HexBytes,
    "eth_chainId": ChainId,
    "eth_getCode": HexBytes,
    "eth_getLogs": List[Log],
    "eth_getBalance": Wei,
    "eth_blockNumber": BlockNumber,
    "eth_accounts": List[Address],
    "eth_getBlockByNumber": Block,
    "eth_getTransactionCount": uint,
    "eth_getTransactionByHash": evmspec.Transaction,
    "eth_getTransactionReceipt": evmspec.FullTransactionReceipt,
    "erigon_getHeaderByNumber": evmspec.ErigonBlockHeader,
    "trace_filter": List[evmspec.FilterTrace],
    "trace_transaction": List[evmspec.FilterTrace],
}
"""
A dictionary mapping RPC method names to their expected return types.
Used to enable more efficient decoding and validation of RPC responses.
"""


_CHAINSTACK_429_ERR_MSG: Final = "You've exceeded the RPS limit available on the current plan."
_QUICKNODE_429_ERR_MSG: Final = (
    # the actual rate limit can vary but the err message always ends like this
    "/second request limit reached - reduce calls per second or upgrade your account at quicknode.com",
)


class PartialResponse(DictStruct, frozen=True, omit_defaults=True, repr_omit_defaults=True):
    """
    Represents a partial JSON-RPC response.

    We use these to more efficiently decode responses from the node.
    """

    result: Raw = None  # type: ignore
    """The result of the RPC call, if successful."""

    error: Optional[Error] = None
    """The error object, if the call failed."""

    @cached_property
    def exception(self) -> Exception:
        """If the rpc response contains an 'error' field, returns a specialized exception for the specified rpc error."""
        if self.error is None:
            raise AttributeError(f"{self} did not error.")

        message = self.error.message
        if "out of gas" in message:
            return OutOfGas(self)
        if "execution reverted" in message or message in {
            "invalid opcode: INVALID",
            "EVM error: InvalidFEOpcode",
        }:
            return ExecutionReverted(self)
        if self.payload_too_large:
            return PayloadTooLarge(self)
        if re.search(r"batch limit (\d+) exceeded", message):
            return ExceedsMaxBatchSize(self)
        if message == "invalid argument 1: hex string without 0x prefix":
            return TypeError(
                message,
                "DANKMIDS NOTE: You're probably passing what should be an integer type as a string type. "
                "The usual culprit is a block number.",
            )
        # chainstack doesnt return a response with status code 429 when we reach rate limits, so we need to handle it specifically here instead of in the usual place
        if _CHAINSTACK_429_ERR_MSG in message:
            return ChainstackRateLimitError(self)
        # quicknode doesnt return a response with status code 429 when we reach rate limits, so we need to handle it specifically here instead of in the usual place
        if message.endswith(_QUICKNODE_429_ERR_MSG):
            return QuiknodeRateLimitError(self)
        return BadResponse(self)

    @property
    def payload_too_large(self) -> bool:
        return any(map(self.error.message.__contains__, constants.TOO_MUCH_DATA_ERRS))

    def to_dict(self, method: Optional[RPCEndpoint] = None) -> RPCResponse:  # type: ignore [override]
        """Returns a complete dictionary representation of this response ``Struct``."""
        data: RPCResponse = {
            key: self.decode_result(method=method, caller=self) if key == "result" else value
            for key, value in self.items()
            if value is not None
        }
        return data

    def decode_result(
        self, method: Optional[RPCEndpoint] = None, *, caller=None
    ) -> Union[HexBytes, Wei, uint, ChainId, BlockNumber, AttributeDict, str]:
        # NOTE: These must be added to the `_RETURN_TYPES` constant above manually
        if method and (typ := _RETURN_TYPES.get(method)):
            if method in (
                "eth_call",
                "eth_blockNumber",
                "eth_getCode",
                "eth_getBlockByNumber",
                "eth_getTransactionReceipt",
                "eth_getTransactionCount",
                "eth_getBalance",
                "eth_chainId",
                "erigon_getHeaderByNumber",
            ):
                try:
                    return better_decode(
                        self.result, type=typ, dec_hook=_decode_hook, method=method
                    )
                except Exception as e:
                    if typ is not Block:
                        raise

                    if e.args[0] == "Object contains unknown field `totalDifficulty`":
                        try:
                            # NOTE should we do this??
                            # _RETURN_TYPES[method] = MinedBlock
                            return better_decode(
                                self.result, type=MinedBlock, dec_hook=_decode_hook, method=method
                            )
                        except ValidationError as e2:
                            if e2.args[0] != "Object contains unknown field `baseFeePerGas`":
                                raise
                            result = better_decode(
                                self.result, type=BaseBlock, dec_hook=_decode_hook, method=method
                            )
                            _RETURN_TYPES[method] = BaseBlock  # all blocks on base are BaseBlocks
                            return result

                    elif e.args[0] == "Object contains unknown field `withdrawals`":
                        typ = ShanghaiCapellaBlock

                    elif e.args[0] == "Object contains unknown field `difficulty`":
                        # I've only seen this on OP stack so far, not sure what difficulty means for their chain
                        typ = BaseBlock

                    else:
                        raise

                    return better_decode(
                        self.result, type=typ, dec_hook=_decode_hook, method=method
                    )

            start = time()
            try:
                decoded = better_decode(self.result, type=typ, dec_hook=_decode_hook, method=method)
            except (ValidationError, TypeError) as e:
                stats.logger.log_validation_error(self, e)
                raise

            if caller:
                stats.log_duration(f"decoding {type(caller)} {method}", start)
            return decoded

        # We have some semi-smart logic for providing decoder hints even if method not in `_RETURN_TYPES`
        if method:
            try:
                if method in _dict_responses:
                    decoded = AttributeDict(decode_nested_dict(self.result))
                    stats.logger.log_types(method, decoded)
                    return decoded
                elif method in _str_responses:
                    # TODO: finish adding methods and get rid of this
                    stats.logger.devhint("Must add `%s: str` to `_RETURN_TYPES`", method)
                    return decode_string(self.result)
            except (ValidationError, TypeError) as e:
                stats.logger.log_validation_error(method, e)

        # In this case we can provide no hints, let's let the decoder figure it out
        decoded = json_decode(self.result)
        if isinstance(decoded, str):
            if method:
                _str_responses.add(method)
            return decoded
        elif isinstance(decoded, dict):
            if method:
                _dict_responses.add(method)
            return AttributeDict(decoded)
        elif isinstance(decoded, list):
            if method is None:
                return decoded
        raise TypeError(
            f"type {type(decoded)} is not supported.  method: {method}  decoded: {decoded}"
        )


class Response(PartialResponse, omit_defaults=True, repr_omit_defaults=True):  # type: ignore [call-arg]
    """
    Represents a complete JSON-RPC response.

    Inherits from PartialResponse and adds the response identifier and JSON-RPC version.
    """

    id: Optional[Union[str, int]] = None
    """The response identifier, matching the request."""

    jsonrpc: Literal["2.0"] = "2.0"
    """The JSON-RPC version, always set to "2.0"."""


def better_decode(
    data: Raw,
    *,
    type: Optional[Type[T]] = None,
    dec_hook: Optional[Callable[[Type, object], T]] = None,
    method: Optional[str] = None,
) -> T:
    decode = _get_decode_func(type, dec_hook)  # type: ignore [arg-type]
    try:
        return decode(data)
    except (ValidationError, TypeError) as e:
        extra_args = [
            f"type: {type.__module__}.{type.__qualname__}",
            f"result: {json_decode(data)}",
        ]
        if method:
            extra_args.insert(0, f"method: {method}")
        e.args = (*e.args, *extra_args)
        raise


@lru_cache_lite_nonull
def _get_decode_func(
    type: Type[T],
    dec_hook: Callable[[Type[T], Union[str, bytes]], T],
) -> Callable[[Union[str, bytes]], T]:
    return Decoder(type=type, dec_hook=dec_hook).decode



from typing import Any, Dict, Final, List, Optional, Tuple, final
from weakref import WeakValueDictionary

from eth_hash import auto

from dank_mids.brownie_patch import _nocompile
from dank_mids.helpers import _helpers


SingletonKey = Tuple[Tuple[str, Any], ...]

_singletons: Final["WeakValueDictionary[SingletonKey, FunctionABI]"] = WeakValueDictionary()


keccak: Final = auto.keccak

_make_hashable: Final = _helpers._make_hashable


# NOTE we must inherit from a python object in order for weakrefs to work, so for now we inherit from _FunctionABI
# https://github.com/python/cpython/blob/9546eeea90c8deb8570a0ef621f075c3c766bc12/Include/internal/pycore_object.h#L874
# https://github.com/python/mypy/blob/501a07b45af8e44eda665e53526fc590dc5a014e/mypyc/irbuild/vtable.py#L17
# https://github.com/python/mypy/blob/501a07b45af8e44eda665e53526fc590dc5a014e/mypyc/codegen/emitclass.py#L285
# TODO: fix the above


@final
class FunctionABI(_nocompile._FunctionABI):
    """
    A class to hold function ABI information.

    This class uses the lru_cache decorator to ensure only one instance is created
    for each unique set of ABI parameters, optimizing memory usage and performance.
    """

    def __init__(self, **abi: Any) -> None:
        """
        Initialize a FunctionABI instance with the given ABI information.

        Args:
            **abi: Keyword arguments representing the ABI of the function.
        """

        self.abi: Final = abi
        """
        The complete ABI (Application Binary Interface) of the function.
        This contains all the information about the function's interface.
        """

        input_sig = build_function_signature(abi)
        self.input_sig: Final = input_sig
        """
        The input signature of the function.
        This is a string representation of the function's parameters and their types.
        """

        self.signature: Final = build_function_selector(input_sig)
        """
        The function selector (4-byte signature) of the function.
        This is used in Ethereum transactions to identify which function to call.
        """

    @staticmethod
    def singleton(**abi: Any) -> "FunctionABI":
        """
        Get a singleton FunctionABI to hold function ABI information.

        This class uses the lru_cache decorator to ensure only one instance is created
        for each unique set of ABI parameters, optimizing memory usage and performance.
        """
        key = tuple((k, _make_hashable(abi[k])) for k in sorted(abi))
        try:
            return _singletons[key]
        except KeyError:
            singleton = _singletons[key] = FunctionABI(**abi)
            return singleton


def get_type_strings(
    abi_params: List[Dict[str, Any]],
    substitutions: Optional[Dict[str, Any]] = None,
) -> List[str]:
    """Converts a list of parameters from an ABI into a list of type strings."""
    types_list = []
    if substitutions is None:
        substitutions = {}

    for i in abi_params:
        type_str: str = i["type"]
        if type_str.startswith("tuple"):
            params = get_type_strings(i["components"], substitutions)
            array_size = i["type"][5:]
            types_list.append(f"({','.join(params)}){array_size}")
        else:
            for orig, sub in substitutions.items():
                if type_str.startswith(orig):
                    type_str = type_str.replace(orig, sub)
            types_list.append(type_str)

    return types_list


def build_function_signature(abi: Dict[str, Any]) -> str:
    types_list = get_type_strings(abi["inputs"])
    return f"{abi['name']}({','.join(types_list)})"


def build_function_selector(input_signature: str) -> str:
    return f"0x{keccak(input_signature.encode()).hex()[:8]}"



import logging
from threading import Thread, current_thread
from typing import Any, Callable, Dict, Tuple

from web3 import Web3
from web3.types import RPCEndpoint

from dank_mids.controller import DankMiddlewareController
from dank_mids.types import AsyncMiddleware

_controllers: Dict[Tuple[Web3, Thread], DankMiddlewareController] = {}

logger = logging.getLogger(__name__)


async def dank_middleware(
    make_request: Callable[[RPCEndpoint, Any], Any], web3: Web3
) -> AsyncMiddleware:
    """
    Creates and returns a DankMiddlewareController middleware function for an asynchronous Web3 instance.

    This function initializes the DankMiddlewareController that batches and processes JSON-RPC
    requests for asynchronous Web3 calls. It is intended to be added to the middleware onion
    of an asynchronous Web3 instance to optimize large-scale blockchain operations.

    Args:
        make_request: The next middleware callable in the chain or the actual request sender.
        web3: The asynchronous Web3 instance to which the middleware will be attached.
            Must be an instance of :class:`~web3.AsyncWeb3`.

    Returns:
        An asynchronous middleware callable that can be used with an asynchronous Web3 instance.

    Examples:
        Setting up Dank Mids middleware with an asynchronous Web3 instance:

        >>> from web3 import AsyncWeb3, HTTPProvider
        >>> from dank_mids.middleware import dank_middleware
        >>> w3 = AsyncWeb3(HTTPProvider("http://localhost:8545"))
        >>> # Inject Dank Mids at the lowest layer so that all calls are batched.
        >>> w3.middleware_onion.inject(dank_middleware, layer=0)
        >>> # Now calling any method uses Dank Mids batching.
        >>> block = await w3.eth.get_block("latest")

        Another example demonstrating the middleware injection:

        >>> from web3 import AsyncWeb3, WebsocketProvider
        >>> from dank_mids.middleware import dank_middleware
        >>> async_w3 = AsyncWeb3(WebsocketProvider("ws://localhost:8546"))
        >>> async_w3.middleware_onion.add(dank_middleware)
        >>> # Retrieve the latest block asynchronously with the middleware active.
        >>> latest = await async_w3.eth.get_block("latest")

    See Also:
        :class:`dank_mids.controller.DankMiddlewareController`: Manages batched JSON-RPC requests.
        :class:`~web3.AsyncWeb3`: The asynchronous Web3 class.
    """
    return DankMiddlewareController(web3)



import itertools
from typing import (
    TYPE_CHECKING,
    Any,
    AnyStr,
    Callable,
    Dict,
    Final,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Union,
    Tuple,
    Type,
    TypeVar,
    final,
    overload,
)

import hexbytes
import msgspec
from eth_abi import decoding, encoding
from eth_abi.abi import default_codec
from eth_abi.encoding import DynamicArrayEncoder, TupleEncoder
from eth_typing import ChecksumAddress, HexStr
from msgspec.json import Decoder, Encoder

if TYPE_CHECKING:
    from dank_mids import types


# due to a circ import issue we will import these later
PartialResponse: Optional[Type["types.PartialResponse"]] = None
Request: Optional[Type["types.Request"]] = None
Response: Optional[Type["types.Response"]] = None
better_decode: Optional[Callable[..., Any]] = None  # type: ignore [type-arg]


__T = TypeVar("__T")


StrEncodable = Union[ChecksumAddress, HexStr]
Encodable = Union[int, StrEncodable, hexbytes.HexBytes, bytes]

RpcThing = Union[HexStr, List[HexStr], Dict[str, HexStr]]


MulticallChunk = Tuple[ChecksumAddress, hexbytes.HexBytes]
MulticallEncoder = Callable[[Tuple[bool, Iterable[MulticallChunk]]], bytes]

DecodedMulticall = Tuple[int, int, Tuple[Tuple["Success", bytes], ...]]
MulticallDecoder = Callable[..., DecodedMulticall]


JSONRPCBatchRequest = List["types.Request"]
# NOTE: A PartialResponse result implies a failure response from the rpc.
JSONRPCBatchResponse = Union[List["RawResponse"], "types.PartialResponse"]
BatchDecoder = Callable[[Union[str, bytes]], Union[List[msgspec.Raw], "types.PartialResponse"]]

# these compile to C constants
HexBytes: Final = hexbytes.HexBytes
Raw: Final = msgspec.Raw
ContextFramesBytesIO: Final = decoding.ContextFramesBytesIO
DecodeError: Final = msgspec.DecodeError

encode_uint_256: Final = encoding.encode_uint_256
decode_string: Final = Decoder(type=str).decode
_decode_raw: Final = Decoder(type=Raw).decode
# due to a forward reference issue we will populate this later
_decode_batch: Optional[BatchDecoder] = None

accumulate: Final = itertools.accumulate
chain: Final = itertools.chain


@final
class RawResponse:
    """
    Wraps a Raw object that we know represents a Response with a `decode` helper method.
    A `RawResponse` is a properly shaped response for one rpc call, received back from a jsonrpc batch request.
    They represent either a successful or a failed response, stored as pre-decoded bytes.
    """

    def __init__(self, raw: msgspec.Raw) -> None:
        self._raw: Final = raw
        """The :class:`Raw` object wrapped by this wrapper."""

    @overload
    def decode(self, partial: Literal[True]) -> "types.PartialResponse": ...
    @overload
    def decode(self, partial: Literal[False] = False) -> "types.Response": ...
    def decode(self, partial: bool = False) -> Union["types.Response", "types.PartialResponse"]:
        """Decode the wrapped :class:`Raw` object into a :class:`Response` or a :class:`PartialResponse`."""
        if better_decode is None:
            __import_from_types()
        return better_decode(self._raw, type=PartialResponse if partial else Response)  # type: ignore [no-any-return, misc]

    __slots__ = ("_raw",)


def decode_raw(data: AnyStr) -> RawResponse:
    """
    Decode json-encoded bytes into a `msgspec.Raw` object.

    Args:
        data: The json-encoded bytes to decode.

    Returns:
        The decoded data wrapped in a RawResponse object.
    """
    try:
        return RawResponse(_decode_raw(data))
    except DecodeError as e:
        e.args = *e.args, data  # type: ignore [has-type]
        raise


def decode_jsonrpc_batch(data: AnyStr) -> Union["types.PartialResponse", List[RawResponse]]:
    """
    Decode json-encoded bytes into a list of response structs, or a single error response struct if applicable.

    Args:
        data: The json-encoded bytes that represent the batch response data.

    Returns:
        Either a PartialResponse if there's an error, or a list of RawResponse objects.
    """
    if _decode_batch is None:
        __make_decode_batch()

    decoded = _decode_batch(data)  # type: ignore [misc]
    return [RawResponse(d) for d in decoded] if isinstance(decoded, list) else decoded


def _encode_hook(obj: Encodable) -> RpcThing:
    """
    A hook function for encoding objects during JSON serialization.

    Args:
        obj: The object to encode.

    Returns:
        The encoded object.

    Raises:
        NotImplementedError: If the object type is not supported for encoding.
    """
    try:
        # We just assume `obj` is an int subclass instead of performing if checks because it usually is.
        return hex(int(obj))  # type: ignore [return-value]
    except TypeError as e:
        # I put this here for AttributeDicts which come from eth_getLogs params
        # but I check for mapping so it can work with user custom classes
        if isinstance(obj, Mapping):
            return {k: _rudimentary_encode_dict_value(v) for k, v in obj.items()}
        else:
            raise TypeError(obj, type(obj)) from e
    except ValueError as e:
        # NOTE: The error is probably this if `obj` is a string:
        # ValueError: invalid literal for int() with base 10:"""
        if isinstance(obj, HexBytes):
            return obj.hex()  # type: ignore [return-value]
        else:
            e.args = *e.args, obj, type(obj)
            raise ValueError(obj, type(obj)) from e


@overload
def _rudimentary_encode_dict_value(value: int) -> HexStr: ...
@overload
def _rudimentary_encode_dict_value(value: __T) -> __T: ...
def _rudimentary_encode_dict_value(value: Union[int, __T]) -> Union[HexStr, __T]:
    # I dont think this needs to be robust, time will tell
    return hex(value) if isinstance(value, int) else value


encode: Final = Encoder(enc_hook=_encode_hook).encode
"""
Encode an object to json-encoded bytes.

This function uses a custom encoding hook defined in `_encode_hook`.

Args:
    obj: The object to encode.

Returns:
    The json-encoded bytes representation of the object.

See Also:
    :class:`msgspec.json.Encoder`: The class that contains underlying JSON encoding function.
"""


# multicall specific stuff

_mcall_encoder: Final[MulticallEncoder] = default_codec._registry.get_encoder(
    "(bool,(address,bytes)[])"
)
_array_encoder: Final[DynamicArrayEncoder] = _mcall_encoder.encoders[-1]  # type: ignore [attr-defined]
_item_encoder: Final[TupleEncoder] = _array_encoder.item_encoder

# We don't need to follow the validation code from eth-abi since we guarantee the input types
_mcall_encoder.validate_value = _array_encoder.validate_value = _item_encoder.validate_value = lambda *_: ...  # type: ignore [attr-defined, method-assign]


def __encode_new(values: Iterable[MulticallChunk]) -> bytes:
    encoded_elements, num_elements = __encode_elements_new(values)
    return encode_uint_256(num_elements) + encoded_elements  # type: ignore [no-any-return]


def __encode_elements_new(values: Iterable[MulticallChunk]) -> Tuple[bytes, int]:
    tail_chunks = [_item_encoder(v) for v in values]
    count = len(tail_chunks)
    head_length = 32 * count
    tail_offsets = chain((0,), accumulate((len(chunk) for chunk in tail_chunks[:-1])))
    head_chunks = (encode_uint_256(head_length + tail_offset) for tail_offset in tail_offsets)
    return b"".join(chain(head_chunks, tail_chunks)), count


_array_encoder.encode = __encode_new  # type: ignore [method-assign]
_array_encoder.encode_elements = __encode_elements_new  # type: ignore [method-assign]


_mcall_decoder: Final[MulticallDecoder] = default_codec._registry.get_decoder(
    "(uint256,uint256,(bool,bytes)[])"
).decode


def mcall_encode(data: Iterable[MulticallChunk]) -> bytes:
    return _mcall_encoder((False, data))


# maybe use this success flag to do something later
Success = bool


def mcall_decode(data: "types.PartialResponse") -> Union[List[bytes], Exception]:
    try:
        decoded = _mcall_decoder(ContextFramesBytesIO(data.decode_result("eth_call")))[2]  # type: ignore [arg-type]
    except Exception as e:
        if PartialResponse is None:
            __import_from_types()
        # NOTE: We need to safely bring any Exceptions back out of the ProcessPool
        e.args = (*e.args, data.decode_result() if isinstance(data, PartialResponse) else data)  # type: ignore [arg-type]
        return e
    else:
        return [tup[1] for tup in decoded]


def __import_from_types() -> None:
    """This helper function is called once to import PartialResponse, Request, Response, and better_decode."""
    global PartialResponse, Request, Response, better_decode
    from dank_mids.types import PartialResponse, Request, Response, better_decode


def __make_decode_batch() -> None:
    from dank_mids.types import PartialResponse

    global _decode_batch
    _decode_batch = Decoder(type=Union[List[msgspec.Raw], PartialResponse]).decode



from contextlib import suppress

from dank_mids._eth_utils import patch_eth_utils
from dank_mids.brownie_patch import (
    DankContractCall,
    DankContractMethod,
    DankContractTx,
    DankOverloadedMethod,
)
from dank_mids.controller import instances
from dank_mids.exceptions import BrownieNotConnectedError
from dank_mids.helpers import setup_dank_w3, setup_dank_w3_from_sync
from dank_mids.middleware import dank_middleware
from dank_mids.semaphores import BlockSemaphore

__all__ = ["dank_middleware", "BlockSemaphore", "setup_dank_w3", "setup_dank_w3_from_sync"]

patch_eth_utils()


def _configure_concurrent_future_work_queue_size() -> None:
    """
    Configures the concurrent futures process pool to allow for a larger number of queued calls.

    This function increases the EXTRA_QUEUED_CALLS value to 50,000, which allows for more
    concurrent operations to be queued in the process pool. This can significantly improve
    performance for applications that make heavy use of brownie.
    """
    import concurrent.futures.process as _cfp

    _cfp.EXTRA_QUEUED_CALLS = 50_000


_configure_concurrent_future_work_queue_size()


# Import brownie objects
__brownie_objects = ["Contract", "dank_web3", "web3", "dank_eth", "eth", "patch_contract"]
with suppress(ImportError):
    from dank_mids.brownie_patch import Contract, dank_eth, dank_web3, patch_contract

    __all__ += __brownie_objects
    # aliased for cleanliness and convenience
    web3 = dank_web3
    eth = dank_eth


def __getattr__(name: str):
    """
    Handles custom attribute access for the 'dank_mids' module.

    This function is called when an attribute that is not found in the module's namespace
    is accessed. If the attribute is one of the brownie-specific objects listed in
    `__brownie_objects`, it raises a `BrownieNotConnectedError`. For all other attributes,
    it raises a standard AttributeError.

    Args:
        name: The name of the attribute being accessed.

    Raises:
        BrownieNotConnectedError: If the attribute is one of `__brownie_objects`.
        AttributeError: If the attribute is not found and is not one of `__brownie_objects`.
    """
    if name in __brownie_objects:
        raise BrownieNotConnectedError(name)
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



# mypy: disable-error-code="attr-defined,dict-item"
from logging import StreamHandler
from typing import Dict, Final

import a_sync
import typed_envs
from a_sync import AsyncProcessPoolExecutor

from dank_mids import _envs
from dank_mids._logging import INFO, getLogger
from dank_mids._mode import OperationMode
from dank_mids.semaphores import BlockSemaphore


logger: Final = getLogger("dank_mids.envs")

if not typed_envs.logger.disabled:
    if not logger.hasHandlers():
        logger.addHandler(StreamHandler())
    if not logger.isEnabledFor(INFO):
        logger.setLevel(INFO)
    logger.info(
        "For your information, you can tweak your configuration for optimal performance using any of the envs below:"
    )


###############
# ENVIRONMENT #
###############

# What mode should dank mids operate in?
# NOTE: infura mode is required for now
# TODO: fix the other modes, set default='default', and make this verbose again
OPERATION_MODE: Final[OperationMode] = _envs.create_env(
    "OPERATION_MODE", OperationMode, default="infura", verbose=False
)  # type: ignore [assignment]

# Max number of eth calls to include in one multicall
MAX_MULTICALL_SIZE: Final = _envs.create_env("MAX_MULTICALL_SIZE", int, default=10_000)
# Max number of rpc calls to include in one batch call
MAX_JSONRPC_BATCH_SIZE: Final = _envs.create_env("MAX_JSONRPC_BATCH_SIZE", int, default=500)
# Maximum amount of requests per second
REQUESTS_PER_SECOND: Final = _envs.create_env("REQUESTS_PER_SECOND", int, default=50)
# Minimum amount of requests per second after rate limit reduction
MIN_REQUESTS_PER_SECOND: Final = _envs.create_env("MIN_REQUESTS_PER_SECOND", int, default=10)

# Enable Demo Mode?
demo_mode: Final = _envs._deprecated_format.create_env(
    "DEMO_MODE", bool, default=False, verbose=False
)
DEMO_MODE: Final = _envs.create_env("DEMO_MODE", bool, default=bool(demo_mode), verbose=False)

# Are you calling a ganache fork? Can't use state override code
ganache_fork: Final = _envs._deprecated_format.create_env(
    "GANACHE_FORK", bool, default=False, verbose=False
)
GANACHE_FORK: Final = _envs.create_env("GANACHE_FORK", bool, default=bool(ganache_fork))
"""Flag indicating whether the current environment is a Ganache fork."""

# We set the default to 20 minutes to account for potentially long event loop times if you're doing serious work.
AIOHTTP_TIMEOUT: Final = _envs.create_env("AIOHTTP_TIMEOUT", int, default=20 * 60)
"""Timeout value in seconds for aiohttp requests."""

# Brownie call Semaphore
#   Used because I experienced some OOM errs due to web3 formatters when I was batching an absurd number of brownie calls.
#   We need a separate semaphore here because the method-specific semaphores are too late in the code to prevent this OOM issue.
brownie_semaphore: Final = _envs._deprecated_format.create_env(
    "BROWNIE_CALL_SEMAPHORE", int, default=100_000, verbose=False
)
BROWNIE_CALL_SEMAPHORE: Final = _envs.create_env(
    "BROWNIE_CALL_SEMAPHORE",
    BlockSemaphore,
    default=brownie_semaphore,
    verbose=not OPERATION_MODE.infura,
)
"""
Semaphore for limiting concurrent Brownie calls.

See Also:
    :class:`dank_mids.semaphores.BlockSemaphore`: The semaphore class used for concurrency control.
"""

BROWNIE_ENCODER_SEMAPHORE: Final = _envs.create_env(
    "BROWNIE_ENCODER_SEMAPHORE",
    BlockSemaphore,
    default=BROWNIE_CALL_SEMAPHORE._default_value * 2,
    verbose=not OPERATION_MODE.infura,
)
"""
Semaphore for limiting concurrent Brownie encoding operations. This limits memory consumption.

See Also:
    :class:`dank_mids.semaphores.BlockSemaphore`: The semaphore class used for concurrency control.
"""

# Processes for decoding. This determines process pool size, not total subprocess count.
# There are 3 pools, each initialized with the same value.
# NOTE: Don't stress, these are good for you and will not hog your cpu. You can disable them by setting the var = 0.
BROWNIE_ENCODER_PROCESSES = _envs.create_env(
    "BROWNIE_ENCODER_PROCESSES",
    AsyncProcessPoolExecutor,
    default=0 if OPERATION_MODE.infura else 1,
    verbose=not OPERATION_MODE.infura,
)
"""
Process pool for Brownie encoding operations.

See Also:
    :class:`a_sync.AsyncProcessPoolExecutor`: The executor class used for managing asynchronous processes.
"""

BROWNIE_DECODER_PROCESSES = _envs.create_env(
    "BROWNIE_DECODER_PROCESSES",
    AsyncProcessPoolExecutor,
    default=0 if OPERATION_MODE.infura else 1,
    verbose=not OPERATION_MODE.infura,
)
"""
Process pool for Brownie decoding operations.

See Also:
    :class:`a_sync.AsyncProcessPoolExecutor`: The executor class used for managing asynchronous processes.
"""

MULTICALL_DECODER_PROCESSES = _envs.create_env(
    "MULTICALL_DECODER_PROCESSES",
    AsyncProcessPoolExecutor,
    default=0 if OPERATION_MODE.infura else 1,
    verbose=not OPERATION_MODE.infura,
)
"""
Process pool for Multicall decoding operations.

See Also:
    :class:`a_sync.AsyncProcessPoolExecutor`: The executor class used for managing asynchronous processes.
"""

COLLECTION_FACTOR: Final = _envs.create_env(
    "COLLECTION_FACTOR",
    int,
    default=10 if OPERATION_MODE.infura else 1,
    verbose=not OPERATION_MODE.infura,
)
"""Factor determining the size of data collection operations."""

# We use a modified version of the request spec that doesn't contain unnecessary fields, and switch to the full spec if necessary for your node.
# Set this env var to any value to force the full request spec always
USE_FULL_REQUEST: Final = _envs.create_env("USE_FULL_REQUEST", bool, default=False, verbose=False)
"""Flag indicating whether to use the full request specification."""

DEBUG: Final = _envs.create_env("DEBUG", bool, default=False, verbose=False)
# NOTE: EXPORT_STATS is not implemented
# TODO: implement this
EXPORT_STATS: Final = _envs.create_env("EXPORT_STATS", bool, default=False, verbose=False)
# NOTE: COLLECT_STATS is implemented
COLLECT_STATS: Final = _envs.create_env(
    "COLLECT_STATS", bool, default=EXPORT_STATS, verbose=not EXPORT_STATS
)

# You probably don't need to use this unless you know you need to
STUCK_CALL_TIMEOUT: Final = _envs.create_env("STUCK_CALL_TIMEOUT", int, default=60 * 2)

# Method-specific Semaphores
method_semaphores: Final[Dict[str, a_sync.Semaphore]] = {
    "eth_call": _envs.create_env(
        "ETH_CALL_SEMAPHORE",
        BlockSemaphore,
        default=BROWNIE_CALL_SEMAPHORE._value,
        verbose=False,
    ),
    "eth_getBlock": _envs.create_env(
        "ETH_GETBLOCK_SEMAPHORE",
        a_sync.Semaphore,
        default=1_000,
        verbose=False,
    ),
    "eth_getCode": _envs.create_env(
        "ETH_GETCODE_SEMAPHORE",
        a_sync.Semaphore,
        default=5_000,
        verbose=False,
    ),
}

if not typed_envs.logger.disabled:
    logger.info("More details can be found in dank_mids/ENVIRONMENT_VARIABLES.py")
    logger.info(
        "NOTE: You can disable these logs by setting the `TYPEDENVS_SHUTUP` env var to any value."
    )


# Validate some stuffs

# NOTE: The other modes are (probably) bugging out right now. More investigation needed. For now you use infura mode.
if not OPERATION_MODE.infura:
    logger.warning(
        "Unless you know what you're doing, dank mids should be run in infura mode for now"
    )

if OPERATION_MODE.infura:
    for process_pool in {
        MULTICALL_DECODER_PROCESSES,
        BROWNIE_DECODER_PROCESSES,
        BROWNIE_ENCODER_PROCESSES,
    }:
        if process_pool._max_workers:
            raise ValueError(
                f"You cannot set env var {process_pool.name} while running dank in infura mode."
            )



from functools import lru_cache
from typing import Callable, Optional

from eth_typing import BlockNumber, ChecksumAddress
from eth_utils import to_checksum_address
from msgspec import Struct
from multicall.constants import MULTICALL2_ADDRESSES

from dank_mids.constants import (
    MULTICALL2_DEPLOY_BLOCKS,
    MULTICALL2_OVERRIDE_CODE,
    MULTICALL3_DEPLOY_BLOCKS,
    MULTICALL3_OVERRIDE_CODE,
)

try:
    from multicall.constants import MULTICALL3_ADDRESSES
except ImportError:
    MULTICALL3_ADDRESSES = {}


class MulticallContract(Struct):
    """
    Represents a multicall contract with its address, deployment block, and bytecode.
    """

    address: ChecksumAddress
    """
    The Ethereum address of the multicall contract.
    """

    deploy_block: Optional[BlockNumber]
    """
    The block number at which the multicall contract was deployed.
    If None, it means the deployment block is unknown.
    """

    bytecode: str
    """
    The bytecode of the multicall contract.
    This is used for state override if necessary.
    """

    needs_override_code_for_block: Callable[[BlockNumber], bool] = None
    """
    A cached function that, when called, determines if the contract needs override code for a specific block.

    Args:
        block: The block number to check.

    Returns:
        True if override code is needed, False otherwise.
    """

    def __post_init__(self) -> None:
        # we don't need to include `self` in the cache key, that's wasteful
        self.needs_override_code_for_block = lru_cache(maxsize=1024)(
            self.__needs_override_code_for_block
        )

    def __needs_override_code_for_block(self, block: BlockNumber) -> bool:
        """
        Determine if the contract needs override code for a specific block.

        Args:
            block: The block number to check.

        Returns:
            True if override code is needed, False otherwise.
        """
        if block == "latest":
            return False
        if self.deploy_block is None:
            return True
        if isinstance(block, str):
            block = int(block, 16)
        return block < self.deploy_block


def _get_multicall2(chainid: int) -> Optional[MulticallContract]:
    if multicall2 := MULTICALL2_ADDRESSES.get(chainid):
        return MulticallContract(
            address=to_checksum_address(multicall2),
            # TODO: copypasta deploy block dict
            deploy_block=MULTICALL2_DEPLOY_BLOCKS.get(chainid),
            bytecode=MULTICALL2_OVERRIDE_CODE,
        )


def _get_multicall3(chainid: int) -> Optional[MulticallContract]:
    if multicall3 := MULTICALL3_ADDRESSES.get(chainid):
        return MulticallContract(
            address=to_checksum_address(multicall3),
            # TODO: copypasta deploy block dict
            deploy_block=MULTICALL3_DEPLOY_BLOCKS.get(chainid),
            bytecode=MULTICALL3_OVERRIDE_CODE,
        )



# NOTE: `FunctionABI` must inherit from a python object in order for weakrefs to work, so for now we inherit from _FunctionABI
# https://github.com/python/cpython/blob/9546eeea90c8deb8570a0ef621f075c3c766bc12/Include/internal/pycore_object.h#L874
# https://github.com/python/mypy/blob/501a07b45af8e44eda665e53526fc590dc5a014e/mypyc/irbuild/vtable.py#L17
# https://github.com/python/mypy/blob/501a07b45af8e44eda665e53526fc590dc5a014e/mypyc/codegen/emitclass.py#L285
# TODO: fix the above


class _FunctionABI: ...
